\ COPYRIGHT 1992@1997 DUTCH FORTH WORKSHOP - FILENAME : ASSWRITE.FRT #0


NEEDS -crc

REVISION -asswrite "-- 2.0 B Start of codespace bug ---"


\ Expects that CodeStart and HeapStart are already defined
\ ******************************************************************************
\ Processor dependant
\ We use that 1 character is 1 address unit
$80000070 VALUE MemStart        \ Lowest usable memory for a T8xx
                                \ Change if compiling for a different machine

4 CONSTANT  CHARS/CELL          \ No! this isn't 1 CELLS !

\ Where we want the workspace to start.
\ The following results in a maximum workspace, but this can be overwritten.
MemStart 5 CELLS + VALUE WorkSpaceStart

\ Where we want the code to start.
$80001000 VALUE CodeStart

\ The workspace length reserved in the assembler for initialised memory
\ The first part, that overlays the boot program, is not used.
$00004000 VALUE WorkSpaceLength
\ The buffer length reserved in the assembler for code
$00008000 VALUE CodeLength

\ We have to skip $100 bytes lest we loose the boot program
\ We use this for fast local variables that cannot be initialised
\ Usage: name ldl,

\ Allocation pointer for fast variables.
\ It starts at 1 because workspace offset zero is used often
\ in various situations
VARIABLE FWP
1 FWP !                      \ The part used for fast variables
: CHECK-FWP FWP @ 64 >= ABORT" Fast WorkSpace overflow" ;
: FWORD   FWP @ CONSTANT   1 FWP +! CHECK-FWP ;    ( "name" N -- )

VARIABLE WP                  \ Used to initialise the workspace
CREATE WorkSpace  WorkSpaceLength ALLOT
64 WP !                      \ The part used for fast variables
: CHECK-WP
    WorkSpaceStart WP @ CELLS + CodeStart >=
      ABORT" Too many Workspace variables: increase CodeStart"
    WP @ 64 + CELLS WorkSpaceLength >=
      ABORT" WorkSpace buffer overflow in assembler"
;
\ Allocate N words(!) in the workspace
: WALLOT  WP +! CHECK-WP ;       ( N -- )
\ Initialise and allocate one word in the workspace
: W,   WorkSpace WP @ CELLS + !   1 WALLOT   ;  ( N -- )
\ Create a named word in the workspace, initialised to N
\ Usage: name ldl,
: WWORD   WP @ CONSTANT   W,   ;    ( "name" N -- )
\ Usage: name dldl,
\ Swap because we are big-endian
: DWORD   WP @ CONSTANT   SWAP W, W,  ;    ( "name" D -- )
\ Store C at position N in the workspace, (so N may be a WWORD )
: WP! WorkSpace SWAP CELLS + ! ;  ( int C, int N -- )
\ Fetch C from position N in the workspace, (so N may be a WWORD )
: WP@ WorkSpace SWAP CELLS + @ ;  ( int N -- int C )

VARIABLE CP        \ Point into CodeSpace, used to assemble
CREATE CodeSpace   CodeLength ALLOT
\ Use only between { and } (!)
: CHECK-CP DP @ CodeSpace CodeLength OVER + WITHIN 0= ABORT" Code space error" ;

\ Make assembly go to the CodeSpace
: {   ALSO ASSEMBLER   CP @ DP @    CP ! DP !  ; ( -- )
\ And back
: }   ONLY FORTH  CHECK-CP   CP @ DP @  CP ! DP !  ; ( -- )
\ Get the current target code address
\ Use only between { and } so HERE is actually CP @
: $    HERE CodeSpace -   CodeStart + ;  ( -- N )
\ Use if not between { and } so we must use CP @
: {$}  CP @ CodeSpace -   CodeStart + ;  ( -- N )

0 VALUE IndentLevel
: PrintName
    CR IndentLevel 2 * SPACES
    $ H. SPACE COUNT TYPE
;
: IncreaseIndentLevel
    1 +TO IndentLevel
;
: LowerIndentLevel
    -1 +TO IndentLevel
;
\ The part between MACRO end END-MACRO is executed (assembled) whenever
\ name is called. Assembly results go to the code space.
: MACRO         ( "name" -- )
    :
    NP @ HEAD>NAME POSTPONE LITERAL POSTPONE PrintName
    POSTPONE IncreaseIndentLevel
    ALSO ASSEMBLER
;
: END-MACRO
    ONLY FORTH
    POSTPONE LowerIndentLevel
    POSTPONE ;
    POSTPONE CHECK-CP
;   IMMEDIATE

\ Heap space, only an allocator
$80008000 VALUE HeapStart \ From here on non-initialised memory
\ HP counts words starting with the address HeapStart
VARIABLE HP   HeapStart HP !
\ Allocate from the heap
: HALLOT  HP +! ;       ( N -- )
\ Create an array with name `name' and length 'N' words on the heap,
\ leave its addres in initialised workspace
\ Usage: name ldl, <offset>, ldnl,
: HARRAY   HP @ WWORD    CELLS HALLOT   ;   ( "name" N -- )

\ Tested
VARIABLE ReturnAddress \ Pass return address offset from PROC to END-PROC
VARIABLE StartAddress  \ Pass start of PROC to RESOLVES
\ Usage: name     ( to execute the procedure  )
\ Return from the current PROC
MACRO return, ( -- )
    ReturnAddress @ ldl, gcall,
END-MACRO

\ Use only within { } , Add prefixes until CP aligns at a word boundary
MACRO align,
    BEGIN
        DP @ 1 CELLS MOD WHILE
        $20 C,                  \ pfix, behaves erroneously
    REPEAT
END-MACRO

\ Between PROC and END-PROC the system is in interpreting mode,
\ because it is simply assembling ( { does that )
ALSO ASSEMBLER
: PROC          ( "name" -- )
    CREATE
        WP @ ,        \ Remember workspace location
        {$} W,        \ Put start address there
        WP @ ReturnAddress ! \ See END-PROC
        CP @ StartAddress !  \ See RESOLVES
        { WP @ stl,   \ Code to save return address one higher
        0 W,        \ Space for return address
        CR ." PROC " NP @ HEAD>NAME COUNT TYPE SPACE $ H.
    DOES>
        DUP BODY> >HEAD HEAD>NAME PrintName ."   ( PROC called)"
        @ ldl,      \ Get procedure start address into A
        gcall,      \ Jump there
;
: END-PROC    return, } ; ( -- )

\ The latest definition is a PROC and resolves the name following.
\ That name better be defined as a PROC !!
\ Essentially both definitions are going to work,
\ using the workspace locations of the old and the code of the new
\ definition. It is mandatory to have this immediately after the PROC
\ definition, because it changes the startup code.
\ Example
\ PROC FORWARD-MONKEY END-PROC
\ PROC MONKEY     RESOLVES FORWARD-MONKEY
\   < eat banana's>
\ END-PROC
: RESOLVES
    ' >BODY              LOCAL OLD      \ Forth data addresses of
    NP @ HEAD>XTA @ >BODY   LOCAL NEW      \  both definitions.
    NEW @ WP@  OLD @ WP!     \ Codestart of NEW, also in old workspace location
    OLD @   NEW !       \ Workspace location of OLD are used
    -2 WALLOT           \ We don't need the WS locations of NEW
    } StartAddress @ CP ! { \ Throw away start code
    OLD @ 1+            \ Workspace index where return address is stored
    DUP ReturnAddress !
    stl,        \ Code to save return address
;

ONLY FORTH

\ ########################## TASK 's ###########################
VARIABLE IN-TASK    FALSE IN-TASK !
ALSO ASSEMBLER
\ Create a task with name "name".
\ Execution of "name" (between { and } ) leads to startup of the
\ task at high prio. No waiting for ending of the task as with la endp, .
\ A is the workspace of the task. The fist idea was to give it the
\ same workspace as the calling process. However that would lead to
\ duplicate use of the workspace[0] location (and possibly others).
\ The workspace starts at W[5], such that there can be no problems
\ with timers alts etc.
\ This intended to make auxiliary high priority tasks to assist
\ one big main calculation program.
\ ldl, and ldlp, are redefined within a TASK such that
\ the locals of the main process can be accessed transparently.
\ The same holds for MACRO 's that only use these instructions,
\ plus workspace independant instructions like ldc,
\ In particular all communication macro's work transparently
\ tested
\ This must use the OLD ldl, instruction!
\ Tested
: TASK ( "name" HARRAY W -- )   \ Should be called outside { }
    CREATE
        {$} ,           \ Remember start address
        ,           \ Remember workspace (the W of Transputer Instruction Set p42)
        {               \ Just start assembling
        TRUE IN-TASK !
        CR ." TASK " NP @ HEAD>NAME COUNT TYPE SPACE $ H.
    DOES>               \ Assume execution of task itself between { }

        DUP @          \ Offset to task start address
        ldc,
        CELL+ @         \ Get TASK workspace (W-MIN-5)
        DUP ldl,
        4 stnl,       \ Where the continuation address is stored

        0 ldlp,         \ Pointer to caller's workspace
        DUP ldl,        \ (W-MIN-5)
        6 stnl,        \ Store pointer at offset 1 of task's workspace

        5 ldc,
        ( W-MIN-5) ldl, wsub,
        runp,           \ At high priority
;

: END-TASK   FALSE IN-TASK !   stopp, } ; ( -- )

\ Wait for two channels, to be used only in a TASK
\ Leaves the channel with input in reg A
\ This must use the OLD ldl, instruction!
\ The result can be fetched using TheMessage
\ This and the following are macro's, but we are already in ASSEMBLER
: ALT2      ( reg A, reg B -- reg A )
    2 LOCAL CH1          \ Offsets in TASK work space
    3 LOCAL CH2
    5 LOCAL TheChannel  \ There is a TheChannel, MACRO that expects 5!
    IN-TASK @ 0= ABORT" Use ALT2 only within tasks "

    CH1 stl, CH2 stl,   \ Remember the channels
    alt,
    CH1 ldl, 1 ldc, enbc,
    CH2 ldl, 1 ldc, enbc,
    altwt,
    CH1 ldl, 1 ldc,  0 ldc, disc, IF, CH1 ldl, TheChannel stl, ENDIF,
    CH2 ldl, 1 ldc,  0 ldc, disc, IF, CH2 ldl, TheChannel stl, ENDIF,
    altend,
    TheChannel ldl,
;

\ Wait for three channels, to be used only in a TASK
\ Leaves the channel with input in reg A
\ This must use the OLD ldl, instruction!
\ The result can be fetched using TheMessage
\ This and the following are macro's, but we are already in ASSEMBLER
\ tested
: ALT3      ( reg A, reg B reg C -- reg A )
    2 LOCAL CH1          \ Offsets in TASK work space
    3 LOCAL CH2
    4 LOCAL CH3
    5 LOCAL TheChannel
    IN-TASK @ 0= ABORT" Use ALT3 only within tasks "

    CH1 stl, CH2 stl, CH3 stl,   \ Remember the channels
    alt,
    CH1 ldl, 1 ldc, enbc,
    CH2 ldl, 1 ldc, enbc,
    CH3 ldl, 1 ldc, enbc,
    altwt,
    CH1 ldl, 1 ldc,  0 ldc, disc, IF, CH1 ldl, TheChannel stl, ENDIF,
    CH2 ldl, 1 ldc,  0 ldc, disc, IF, CH2 ldl, TheChannel stl, ENDIF,
    CH3 ldl, 1 ldc,  0 ldc, disc, IF, CH3 ldl, TheChannel stl, ENDIF,
    altend,
    TheChannel ldl,
;

: TheChannel, IN-TASK @ 0= ABORT" Use only within tasks " 5 ldl, ;

DEFINITIONS     \ The following to be added to the assembler wordlist
\ tested (all 3)
: stl,  IN-TASK @ IF 1 ldl, stnl,      ELSE stl,  THEN ;
: ldlp, IN-TASK @ IF ldc, 1 ldl, wsub, ELSE ldlp, THEN ;
\ Redefine last! It is used by the previous words!
: ldl,  IN-TASK @ IF 1 ldl, ldnl,      ELSE ldl,  THEN ;
ONLY FORTH    DEFINITIONS
\ tested
: CHANNEL, MININT WWORD ;

VARIABLE DUMMY
: write-cell ( word f -- )
    >R DUMMY !
    DUMMY CHARS/CELL R> WRITE-FILE  \ Yes 4 not 1 CELLS we need characters here
    ABORT" writing of a cell went wrong"
;

\ Data to communicate the memory lay out
\ They are all are initialised before the program is written to file
\ but just now they are not all known.
\ The last three are the first three words of the resulting .x file
0        WWORD I_BootStart      \ Start of boot program
0        WWORD I_BootLen DECIMAL \ Probably too large
0        WWORD I_StartArea      \ Start of initialised mem area
0        WWORD I_AreaLen        \ Length of initialised mem area
0        WWORD I_EntryPoint     \ Code execution starts at this address
0        WWORD I_CodeStart      \ Start of code (in behalf of CRC)
0        WWORD I_CodeEnd        \ End of code (in behalf of CRC)
7 CONSTANT #SPECIAL-LOCATIONS

: CrcAlign      \ Align and add a crc check to the code

    \ Align
    BEGIN
       CP @ 1 CELLS MOD WHILE
       0 CP @ C!
       1 CP +!
    REPEAT

    \ Add crc to code.
    CodeSpace   CP @   CrcOfArea
   ( Crc) CP @ !   1 CELLS CP +!

   \ Remember end of code area
   {$} I_CodeEnd WP!
;

\ Write the current codespace and program to a file "name"
\ tested
: WriteOut      ( "name" -- )
    0 LOCAL binfile
    CrcAlign
    WorkSpaceStart 256 +   I_StartArea   WP!   \ Start of initialised mem area
    {$}   I_StartArea WP@   -   I_AreaLen WP! \ Length of initialised mem area

    CR ." Writing executable..." CR
    BL PARSE W/O BIN CREATE-FILE  ABORT" writepr: Unable to create file"
                TO binfile
    I_AreaLen WP@    binfile write-cell        \ size
    I_StartArea WP@  binfile write-cell
    I_EntryPoint WP@ binfile write-cell
    \ Write our assembly time copy of the workspace, except the
    \ initialised first part.
    WorkSpace 256 + CodeStart WorkSpaceStart 256 + - binfile WRITE-FILE
    ABORT" writepr: unable to write file"
    \ Write the assembled code, no use going farther than CP @
    CodeSpace   CP @   OVER - binfile WRITE-FILE
    ABORT" writepr: unable to write file"
    binfile CLOSE-FILE
    ABORT" writepr: unable to close file"
;

\ Write the current codespace as a boot program to a file "name"
\ The length to be sent is derived from the file length bye the
\ booter such that no header is needed in this PriteOut file
\ If the booter already exists is only grows (bug in CREATE-FILE)
\ tested
: WriteBoot ( "name" -- )
    0 LOCAL binfile
    CrcAlign
    CR ." Writing bootfile"
    BL PARSE W/O BIN CREATE-FILE  ABORT" writepr: Unable to create file"
                TO binfile
    \ Write the assembled code, no use going farther than CP @
    CodeSpace   CP @   OVER - binfile WRITE-FILE
        ABORT" writepr: unable to write file"
    binfile CLOSE-FILE
        ABORT" writepr: unable to close file"
;

\ Add the boot program from file "name" to the CodeSpace.
\ Remember its length and start in the special locations.
: AddBoot ( "name" -- )
   0 LOCAL FID          \ File id of boot file
   {$} I_BootStart WP!     \ Start of boot program

   \ Find booter size.
   BL PARSE R/O BIN OPEN-FILE  ABORT" No boot file?"   TO FID
   {   \ Start assembling
       HERE 256 FID READ-FILE
       ABORT" Asswrite: can't read boot program."
       DUP   2 $FF WITHIN 0= ABORT" Wrong size for a boot program"
       DUP ALLOT
   }
   I_BootLen WP!            \ Fill in boot program's size

   FID CLOSE-FILE ( return code) DROP
;

: ENTER-HERE    $ I_EntryPoint WP! ;

\ #################### General Space management ###################

\ Get things in order, after the user adjusted
\ Heapstart CodeStart WorkspaceStart to his needs.
\ Will become much more complicated if the assembler is going to
\ reallocate its buffers if it cannot cope with the addresses wanted.
: INIT-ASSEMBLER

   \ Zero the Workspace again
   WorkSpace  WorkSpaceLength $E5 FILL
   CodeSpace  CodeLength      $5E FILL

   1 FWP !
   64 #SPECIAL-LOCATIONS + WP !
   HeapStart HP !
   CodeSpace CP !

   \ Fill in special locations
   MemStart I_BootStart WP!     \ Start of boot program
   \ Find booter size.
   S" myboot.x" R/O BIN OPEN-FILE  ABORT" No boot file?"
   DUP FILE-SIZE ABORT" No boot file size?"
   ROT CLOSE-FILE ( return code) DROP
   D>S I_BootLen WP!            \ Fill in boot program's size
   {$} I_CodeStart WP!          \ Program start for CRC-check.

   $00004000 WorkSpaceLength <>
    ABORT" You cannot change the buffer length for the workspace, yet"
   $00008000 CodeLength <>
    ABORT" You cannot change the buffer length for the code, yet"
;

\ Check the memory usage from low to high memory.
: Checks
    \ Check fast workspace.
    64   FWP @ -   CR . ." spare words in the fast (unitialised) workspace area)"
    CodeStart   WorkSpaceStart WP @ CELLS + -
    CR . ." spare bytes over in the normal workspace area"
    HeapStart CodeStart -   CP @ CodeSpace -    -
    CR . ." spare bytes for the code"
    HP @    HeapStart  -  CR . ." bytes used for heap"
    \ Wrong
    MININT $100000 +    HP @   -   1000000 -
    CR . ." bytes left on a 1M machine "
    MININT $400000 +    HP @   -   4000000 -
    CR . ." bytes left on a 4M machine "
;

\ Init, using reasonable defaults.
INIT-ASSEMBLER
