\ COPYRIGHT 1995@2000 DUTCH FORTH WORKSHOP - FILENAME : commun1.FRT #0

REVISION -commun1 "2.0B -- commun1.frt : using PrimeLen ---"

\ #################### processor configuration #####################

\ This processor configuration is told us by the master boot link

0 WWORD ChainLen     \ #transputers hanging off us,
                     \ a negative value indicates a problem.
\ ChainLen also identifies the processor while debugging
\ Communicate this address to the communication module.
ChainLen TO ProcessorId

\ #################### Booting & Starting up ################################
PRIVATES
\ Signal error in booting by negating ChainLen. This is about the only
\ thing that might get through. Map N to -1-N in order to be able to signal
\ an error also if N were 0.
MACRO BootError
    -1 ldc, ChainLen ldl, sub, ChainLen stl,
END-MACRO PRIVATE

\ Send our own boot program to the down link as a boot program
\ It must not have been destroyed, by use of FWORD 's!
MACRO Boot
    I_BootLen BS->,             \ Send length (a byte)
    I_BootStart I_BootLen $S->, \ Send Boot program
END-MACRO PRIVATE

\ Send our own program down the link.
MACRO SendProgram
    I_AreaLen   WS->,
    I_StartArea WS->,
    I_EntryPoint WS->,
    I_StartArea I_AreaLen   $S->,
END-MACRO PRIVATE

\ This sends essentially the same message down as we recieved ourselves
MACRO SendInit
    %>init   MS->,    SieveSize WS->,  ChainLen WS->,
    \ Down transputer is booting through
    %>init MR<-,      MessageRight WR<-,    \ Get their chain len
        MessageRight ldl, 0<, IF,
            \ It was wrong already, prepare to send through
            MessageRight ldl, ChainLen stl,
        ELSE,
            MessageRight ldl, 1 adc, ChainLen ldl, <>, IF,
                \ New problem, try to get through that we identified
                \ a problem.
                BootError
            ENDIF,
        ENDIF,
END-MACRO PRIVATE

MACRO BootThrough

    ChainLen ldl, 0 ldc, >, IF,
        Boot
        MessageRight WR<-,      \ ignore TID
        SendProgram
        SendInit
    ENDIF,

END-MACRO PRIVATE

PROC    BootChain

    \ Wait for command
    %>init   MR->,    SieveSize WR->,  ChainLen WR->,
    ChainLen ldl, -1 adc, ChainLen stl,  \ We have one less than our master
        I_CodeStart ldl,   -1 ldc, I_CodeEnd ldl, wsub,   (CheckCrc) IF,
          BootError    \ Program corrupted
    ELSE,
        BootThrough   \ Boot down chain
    ENDIF,

    \ Notify that all went okay or indicate a problem by a negative ChainLen
    %>init MS<-,      ChainLen WS<-,

END-PROC

PROC CheckCrc
    I_CodeStart ldl,
    -1 ldc, I_CodeEnd ldl, wsub,
   (CheckCrc) IF,
        18 ERROR,
        p ldl, DEBUG,
    ENDIF,
END-PROC


DEPRIVE

\ #################### Accumulation of %<twin messages #####################

\ ************************* Status ************************************

\ Whenever WaitingForBatch is set, the offset in Primebuffer reflects
\ what is about to arrive and the buffer itself is invalid.
0 WWORD CumulTwin \ Where twin count is accumulated PRIVATE

\ The 1 in WaitingForBatch blocks the request for the first buffer
\ until there has been a request to start sieving.
1 WWORD WaitingForBatch     \ A request for a batch has been sent up
0 WWORD DownRequested       \ The down chain requests a batch
0 WWORD CalcRequested       \ The calculation requests a batch.
0 WWORD NextDownRequested       \ The down chain requests a batch
0 WWORD NextCalcRequested       \ The calculation requests a batch.
0 WWORD CalcReady           \ The local calculation is ready
0 WWORD DownReady           \ The down chain is ready with its calculation
0 WWORD ReservedForDown     \ The buffer is reserved for the down chain
0 WWORD ReservedForCalc     \ The buffer is reserved for the calculation
0 WWORD PingRequested       \ A request has been placed but not executed
                            \ for information.

\ Set and clear a flag, e.g. as above.
MACRO SET,                      ( WWORD FLAG -- )
    1 ldc, stl,
END-MACRO

MACRO CLEAR,                    ( WWORD FLAG -- )
    0 ldc, stl,
END-MACRO

\ Add the number of numbers in a primebatch to A
MACRO PrimeLenAdd,              ( regA -- regA )
    PrimeLen ldl,
    PPBY ldc, mul,
    add,
END-MACRO

CHANNEL, NOTIFY-UP
CHANNEL, NOTIFY-DOWN


\ #################### The main communication task that sends up ######################################

PRIVATES
0. DWORD ToUpMessage PRIVATE     \ Message such as recieved by us, possibly double
0 WWORD MessageProc PRIVATE      \ The Processor id that is recieved together with
                        \ DEBUG, and ERROR, messages, and send through.

\ Send the information requested by ping.
MACRO DoPing
         p ldl, DEBUG,
         CommPrimes ->Offset ldl, DEBUG,

         0 ldc,
         WaitingForBatch     ldl,  or, 1 ldc, shl,
         DownRequested       ldl,  or, 1 ldc, shl,
         CalcRequested       ldl,  or, 1 ldc, shl,
         NextDownRequested   ldl,  or, 1 ldc, shl,
         NextCalcRequested   ldl,  or, 1 ldc, shl,
         CalcReady           ldl,  or, 1 ldc, shl,
         DownReady           ldl,  or, 1 ldc, shl,
         ReservedForDown     ldl,  or, 1 ldc, shl,
         ReservedForCalc     ldl,  or,
         DEBUG,
         PingRequested CLEAR,
END-MACRO


\ Awake the down sending such that it may send message that are
\ available.
\ Note that this and the other task are high priority and hence run to
\ completion. This means that there is no race condition with respect to
\ the content of the NOTIFY-DOWN.
MACRO NotifyDown
      NOTIFY-DOWN ldl, mint, <>, IF, \ If waiting
        NOTIFY-DOWN ldlp, %>notify ldc, outword,
        \ Else not waiting, i.e. sending a message.
        \ In that case it will see there is work anyway.
      ENDIF,
END-MACRO PRIVATE

\ Request a batch with the current offset `Offset'
MACRO RequestBatch
    %<requestbatch MS<-,   CommPrimes ->Offset WS<-,
    WaitingForBatch SET,
END-MACRO PRIVATE


\ The calculation process requests a primebatch with offset `Offset'
MACRO CalculationRequestsBatch  ( WWORD Offset -- )
    LOCAL Offset
    Offset ldl, CommPrimes ->Offset ldl, =, IF,
        CalcRequested  SET,
        ReservedForCalc ldl, IF,
            WaitingForBatch ldl, IF,   #24 ERROR, ENDIF,
        ENDIF,
    ELSE, CommPrimes ->Offset ldl, PrimeLenAdd,   Offset ldl,   =, IF,
        NextCalcRequested  SET,
    ELSE,
        #132 ERROR,
    ENDIF,
    ENDIF,
END-MACRO PRIVATE

\ The down chain requests a primebuffer with offset in Offset
MACRO DownChainRequestsBatch  ( WWORD Offset -- )
    LOCAL Offset
    Offset ldl, CommPrimes ->Offset ldl, =, IF,
        DownRequested  SET,
        ReservedForDown ldl, IF,
            WaitingForBatch ldl, IF,   #24 ERROR, ENDIF,
        ENDIF,
    ELSE, CommPrimes ->Offset ldl, PrimeLenAdd,   Offset ldl,   =, IF,
        NextDownRequested  SET,
    ELSE,
        #131 ERROR,
    ENDIF,
    ENDIF,
END-MACRO PRIVATE

MACRO do%<twin
    TheChannel, COMM<-COMP ldlp, <>, IF,
        DownReady       SET,
        ReservedForDown CLEAR,
    ELSE,
        CalcReady       SET,
    ENDIF,
    CumulTwin ldl, ToUpMessage ldl, add, CumulTwin stl,
END-MACRO PRIVATE


COMM-WS-UP TASK COMM-UP
BEGIN,
    ?ERROR,                   \ Send error if present

    PingRequested ldl, IF,
        DoPing
    ENDIF,

    WaitingForBatch ldl, ReservedForCalc ldl, or, ReservedForDown ldl, or,
    0=, IF,
        NextCalcRequested ldl,   CalcRequested stl,
        NextCalcRequested CLEAR,
        NextDownRequested ldl,   DownRequested stl,
        NextDownRequested CLEAR,
        CommPrimes ->Offset ldl,   BPPB PPBY * adc,   CommPrimes ->Offset stl,
        RequestBatch
    ENDIF,

    \ Give the other communication process a chance to react on state change

    \ Wait on all inputs at the same time
    NOTIFY-UP ldlp,
    COMM<-COMP ldlp,
    LINKDOWN-IN ldc,
    ALT3 ToUpMessage WR,
\     ALT2 ToUpMessage WR,

    \ Message dispatcher

    \ Just a message that awakens us
    ToUpMessage ldl, %>notify ldc, =, IF,
        \ Do nothing

    \ Results are flowing in
    ELSE, ToUpMessage ldl, %<twin ldc, =, IF,
        TheChannel, ToUpMessage WR,
        do%<twin
        CalcReady ldl, DownReady ldl, and, IF,
            %<twin MS<-, CumulTwin WS<-,
        ENDIF,

    \ Send through a request for a prime buffer
    ELSE, ToUpMessage ldl, %<requestbatch ldc, =, IF,
        TheChannel, ToUpMessage WR,   \ Recieve offset
        TheChannel, COMM<-COMP ldlp, =, IF,
            ToUpMessage CalculationRequestsBatch
        ELSE,
            ToUpMessage DownChainRequestsBatch
        ENDIF,

    \ Send through the reflection of where this transputer starts.
    ELSE, ToUpMessage ldl, %<showstart ldc, =, IF,
        TheChannel, MessageProc WR,
        TheChannel, ToUpMessage DR,
        %<showstart MS<-,   MessageProc WS<-,   ToUpMessage DS<-,

    \ Send through any debug messages
    ELSE, ToUpMessage ldl, %<debug ldc, =, IF,
        TheChannel, MessageProc WR,
        TheChannel, ToUpMessage WR,
        %<debug MS<-,   MessageProc WS<-,   ToUpMessage WS<-,

    \ Send through any error messages
    ELSE, ToUpMessage ldl, %<error ldc, =, IF,
        TheChannel, MessageProc WR,
        TheChannel, ToUpMessage WR,
        %<error MS<-,   MessageProc WS<-,   ToUpMessage WS<-,

    \ Protocol error
    ELSE,
        1 ERROR,
        ToUpMessage ldl, DEBUG,      \ This prevents an avalanche of errors
    ENDIF,
    ENDIF,
    ENDIF,
    ENDIF,
    ENDIF,
    NotifyDown
    ENDIF,
AGAIN,
END-TASK
DEPRIVE

\ #################### The main communication task that sends down ######################################

PRIVATES

0. DWORD ToDownMessage     \ Message such as recieved by us, possibly double
0 WWORD MessageProc     \ The Processor id that is recieved together with
                        \ DEBUG, and ERROR, messages, and send through.
\ See also the explanation of these numbers in twin.frt.
\ The meaning of these is similar.
0. DWORD ComNumberStart PRIVATE    \ Contains start for whole down chain.
0. DWORD ComNumberLen   PRIVATE    \ Contains Length for whole down chain.


\ Remember the number `n' as an error.
\ It will be sent at the first opportunity.
MACRO ERROR,            ( n -- )
    ldc, ERROR stl,
END-MACRO PRIVATE

\ Initalise the whole communication
MACRO InitCom
    0 ldc, CumulTwin stl,
    DownRequested   CLEAR,
    CalcRequested   CLEAR,
    NextDownRequested   CLEAR,
    NextCalcRequested   CLEAR,
    CalcReady       CLEAR,
    ReservedForDown CLEAR,
    ReservedForCalc CLEAR,

    \ Fake the completion of handling the -1 batch
    WaitingForBatch CLEAR,
    0 ldc, PrimeLenAdd, neg,   CommPrimes ->Offset stl,
END-MACRO

\ Awake the up sending such that it may send message that are
\ available.
\ Note that this and the other task are high priority and hence run to
\ completion. This means that there is no race condition with respect to
\ the content of the NOTIFY-UP.
MACRO NotifyUp
      NOTIFY-UP ldl, mint, <>, IF, \ If waiting
        NOTIFY-UP ldlp, %>notify ldc, outword,
        \ Else not waiting, i.e. sending a message.
        \ In that case it will see there is work anyway.
      ENDIF,
END-MACRO PRIVATE

\ Handle the start message. Do as much as possible and send
\ the remainder down.
MACRO do%>start
    InitCom
    \ We do the last part of the sieve
    ComNumberLen SieveSize DS-, ComNumberLen dstl,

    DownReady       SET,    \ Default: do nothing down.
    ComNumberLen 1+ ldl, -1 ldc, >, IF,        \ Not negative
        ComNumberLen ldl, ComNumberLen 1+ ldl, or, IF,    \ Not zero
            DownReady       CLEAR,
        ENDIF,
    ENDIF,

    \ Now send command down if needed, leaving into NumberStart/NumberLen
    \ what we have do ourselves.
    DownReady ldl, IF,
        \ We do everything ourselves.
        \ Correct by adding SieveSize back up.
        ComNumberLen SieveSize DS+, ComNumberLen dstl,
    ELSE,
        \ Send remainder down, that is currently in NumberStart/NumberLen

        \ If no one listens, that is an error.
        ChainLen ldl, 0=, IF, 2 ERROR, ENDIF,
        %>start MS->, ComNumberStart DS->,   ComNumberLen DS->,
        MASK-COPY MaskLen $S->,

        \ We ourselves are doing a full sieve, the topmost part.
        \ Sieving NumberStart[NumberLen, NumberLen+SieveSize]
        \ NumberStart is adjusted to the start of the table
        ComNumberStart ComNumberLen ComNumberStart DD+, \ 64-bit
        SieveSize dldl, ComNumberLen dstl,
    ENDIF,

    \ Start our own calculation task
    %>start MSM->P,   ComNumberStart DSM->P,   ComNumberLen DSM->P,

END-MACRO


\ Send to the calculation
MACRO SendToCalc ( -- )
    %>primebatch MSM->P,
    CommPrimes ->Offset WSM->P,
    CommPrimes ->Buffer   PrimeLen   $SM->P,
    CalcRequested   CLEAR,
    ReservedForCalc CLEAR,
END-MACRO

MACRO SendDown   ( -- )
    %>primebatch MS->,
    CommPrimes ->Offset WS->,
    CommPrimes ->Buffer   PrimeLen   $S->,
    DownRequested  CLEAR,
    ReservedForDown CLEAR,
END-MACRO

\ Recieve a prime buffer message.
MACRO RecievePrimeBatch   ( WWORD Offset -- )
    LOCAL Offset

    Offset ldl,  CommPrimes ->Offset ldl, <>, IF,
        7 ERROR,
    ENDIF,
    WaitingForBatch ldl, 0=, IF,   24 ERROR, THEN,
    CommPrimes ->Buffer   PrimeLen   $R->,
    WaitingForBatch CLEAR,
    ReservedForCalc SET,
    1 ldc, DownReady ldl, sub, ReservedForDown stl,

    CalcRequested ldl, IF,
        SendToCalc
    THEN,

    DownRequested ldl, IF,
        SendDown
    THEN,

END-MACRO


COMM-WS-DOWN TASK COMM-DOWN
BEGIN,

    \ Wait on up link
    \ and on a possible notification
    NOTIFY-DOWN ldlp,
    LINKUP-IN ldc,
    ALT2 ToDownMessage WR,

    \ Check this out for each change in state
    WaitingForBatch ldl, 0=, IF,
        ReservedForCalc ldl, CalcRequested ldl, and, IF,
            SendToCalc
        ENDIF,
        ReservedForDown ldl, DownRequested ldl, and, IF,
            SendDown
        ENDIF,
    ENDIF,

\      Message dispatcher
    \ They  awaken us, probably data available
    ToDownMessage ldl, %>notify ldc, =, IF,
        \ Do nothing

    \ Start a counting operation in cooperation with the other
    \ transputers of this chain.
    ELSE, ToDownMessage ldl, %>start ldc, =, IF,
        \ Get remainder of message
        ComNumberStart DR->,   ComNumberLen DR->,   MASK-COPY MaskLen $R->,
        do%>start

    \ Recieve a requested prime buffer
    ELSE, ToDownMessage ldl, %>primebatch ldc, =, IF,
        TheChannel, ToDownMessage WR,   \ Recieve offset
        ToDownMessage RecievePrimeBatch

    \ Show what is currently going on.
    \ This message is most useful if it hangs...
    ELSE, ToDownMessage ldl, %>ping ldc, =, IF,
        PingRequested SET,      \ The upsending process must do this.
        ProcessorId ldl, IF, %>ping MS->, ENDIF,

    \ Protocol error
    ELSE,
        1 ERROR,
        ToDownMessage ldl, DEBUG,      \ This prevents an avalanche of errors
    ENDIF,
    ENDIF,
    ENDIF,
    ENDIF,
    NotifyUp
AGAIN,
END-TASK
DEPRIVE
