\ COPYRIGHT 1993 DUTCH FORTH WORKSHOP - FILENAME : fileloader
\ LANGUAGE    : ANS Forth with environmental dependancies
\ PROJECT     : tForth
\ DESCRIPTION : formerly t-Server
\ CATEGORY    : Utilities
\ AUTHOR      : Albert van der Horst (with thanks to Erwin Dondorp)
\ LAST CHANGE : 970221 AH FCR0073 The great rename, applied to meta source and application
\ LAST CHANGE :
\


DOC

 This is a program that loads an assembly level executable
 via the standard boot program to a link.

 This code has worked with a Parsytec TPM-PC boards, connected via
 LINK1 to an Atari "workstation" board, where the reset was done
 by the non-inverting output of the tmp-pc only pulling down
 through a diode.

ENDDOC

        ?DEF NEEDS 0= [IF] INCLUDE NEEDS.FRT [THEN]
        NEEDS    -miscutil
        REVISION -filo "--- file loader Version 0.03 ---"

        PRIVATES


0 VALUE networktype                            \ 0 unspecified
                                               \ 1 CSA
                                               \ 2 Parsytec

0 VALUE bootfileid              PRIVATE
0 VALUE port
0 VALUE use-in                  PRIVATE
0 VALUE use-out                 PRIVATE
CREATE filename  0 C, 255 ALLOT PRIVATE
0 VALUE binfile                 PRIVATE
FALSE VALUE bootwanted
FALSE VALUE resetwanted
CREATE cmdl 0 C,   #1000 ALLOT  PRIVATE
0 VALUE cmdp                    PRIVATE \ feeds the character of cmdl one by one
CREATE tbuffer    #10000 ALLOT  PRIVATE
CREATE bootfiletext #256 ALLOT  PRIVATE
CREATE codeinfo  5 CELLS ALLOT  PRIVATE

: reset-options
    FALSE TO bootwanted
    FALSE TO resetwanted
    S" prog.x" filename $!
    0 0 cmdl $!
;

\ Rely on proper path / append

CREATE tfboot$  ," myboot.x"    PRIVATE


\ --------- RESET SPECIFIC FOR CSA -----------------------------------------

        0 =: 'Reset-c   PRIVATE \ Reset the [down linked] transputers here

\ Reset everything hanging off this transputer
: RESET-LINK-C
        0 'Reset-c C!   #50 MS          \ make sure reset is off,
        1 'Reset-c C!   #50 MS          \ assert for 50 ms,
        0 'Reset-c C!                   \ deassert forever.
        ; PRIVATE

\ --------- RESET SPECIFIC FOR PARSYTEX -----------------------------------------
\
$000000C0 =: 'Reset-p   PRIVATE \ for Parsytec TPM-PC


\ Reset the links hanging of a Parsytec board
: RESET-LINK-P ( -- )
        \ Send a special code to the reset address to activate the PAL
        4 0 DO I 'Reset-p ! LOOP
        \ Find bit mask belonging to the output link
        use-in
        CASE
         LINK0-IN OF 1 ENDOF
         LINK1-IN OF 2 ENDOF
         LINK2-IN OF 4 ENDOF
         LINK3-IN OF 8 ENDOF
         TRUE ABORT" No port number given"
        ENDCASE
        'Reset-p !
        100 MS              \ recommended: 2 ticks = 128 usec.
        0 'Reset-p !
        10 MS
        use-in reset-channel drop
        use-out reset-channel drop
; PRIVATE


\ DETERMINE ---------------------------------------------------

: sendbooter
        tfboot$ $@ R/O BIN OPEN-FILE SWAP TO bootfileid
        IF TRUE  ." T-Server: can't find or open " tfboot$ $@ TYPE ABORT THEN
        bootfiletext #256 bootfileid READ-FILE
                  ABORT" T-Server: can't read tfboot.x"
        DUP  0=   ABORT" tfboot.x cannot be 0 bytes"
        DUP  1  = ABORT" tfboot.x cannot be 1 byte"
        DUP $FF > ABORT" tfboot.x cannot be larger than 255 bytes"
        CR ." Sending booter"
        {{
           DUP use-out CHANNEL-C!
           bootfiletext SWAP use-out CHANNEL-SEND
        }}
        bootfileid CLOSE-FILE ABORT" cannot close file tfboot.x"
; PRIVATE

: determine
        SELECT
          use-in GUARD ENDGUARD
          #1000 >TICKS TIMEOUT
             TRUE ABORT" filo: time out in waiting for magic number"
          ENDGUARD
        ENDSELECT
        use-in CHANNEL-@  DUP .
        -30 <> ABORT" filo: wrong magic number"
; PRIVATE


\ SEND a block of the tforth executable

: sendfileblock
        LOCAL l
        0 LOCAL rl

        BEGIN   l 0>
        WHILE   l #10000 MIN TO rl
                tbuffer rl binfile READ-FILE ABORT" T-Server: Read error on file"
                        TO rl

                tbuffer rl use-out CHANNEL-SEND

                rl NEGATE +TO l
        REPEAT
; PRIVATE

: sendnulls     0 DO 0 use-out CHANNEL-C! LOOP ; PRIVATE

: sendexecutable
        CR ." Sending executable"
        filename $@ R/O BIN OPEN-FILE  ABORT" T-Server: Unable to open file"
                TO binfile

        codeinfo 3 CELLS binfile READ-FILE ABORT" T-Server: Unable to read from file"
        3 CELLS <> ABORT" T-Server: binary file too short"


     {{
        \ total size
        codeinfo 0 CELLS + @  use-out CHANNEL-!  \ size
        codeinfo 1 CELLS + @  use-out CHANNEL-!  \ memory start and execution start
        codeinfo 2 CELLS + @  use-out CHANNEL-!  \ workspace start
        codeinfo 0 CELLS + @  sendfileblock      \ code block
     }}
        binfile CLOSE-FILE ABORT" T-Server: Unable to close file" ; PRIVATE


: doreset
        resetwanted IF
            ." Resetting transputer..." CR
            networktype
            CASE    1 OF RESET-LINK-C ENDOF
                    2 OF RESET-LINK-P ENDOF
                    TRUE ABORT" T-Server: specify networktype using /s"
            ENDCASE
        ENDIF
 ; PRIVATE

: doboot
        bootwanted IF
            ." Booting..." CR
            sendbooter
            determine
            sendexecutable
        ENDIF
 ; PRIVATE


\ Find and checks the addresses of the in and output links to be used
: finduse               ( -- )
        port
        CASE
         0 OF LINK0-IN TO use-in LINK0-OUT TO use-out ENDOF
         1 OF LINK1-IN TO use-in LINK1-OUT TO use-out ENDOF
         2 OF LINK2-IN TO use-in LINK2-OUT TO use-out ENDOF
         3 OF LINK3-IN TO use-in LINK3-OUT TO use-out ENDOF
        ENDCASE
        use-in BOOTLINK-IN =
        ABORT" Cannot use own boot link to boot another transputer"
; PRIVATE

\ Auxiliary word to parse an inline numeric argument
\ optionally ending in K or M
: parsenumber    ( "number" -- <n> )
        0.0 BL PARSE >NUMBER       \ d c-addr u
        CASE
                0 OF DROP FALSE ENDOF
                1 OF C@ >UPPER
                        DUP [CHAR] K = IF
                                DROP D>S 1024 UM* FALSE
                        ELSE [CHAR] M = IF
                                D>S $100000 UM* FALSE
                        ELSE
                                TRUE
                        ENDIF ENDIF
                ENDOF
                TRUE SWAP       \ Swap the to be discarded number onto the top
        ENDCASE

        ABORT" This is not a number, as needed with the previous option!"
        D>S
; PRIVATE

: /b    TRUE TO bootwanted ;
: /r    TRUE TO resetwanted ;
: /s    parsenumber TO networktype ;
: /p    parsenumber TO port ;

: /f    BL PARSE filename $! ;                    \ "filename" --- <>

: /?    CR ." filo"
        CR ." loader for assembler programs"
        CR ." options compatible with server `tforth' except /g/m missing"
        CR ." `port' `resetwanted' `bootwanted' `networktype' direct accessable"
;

: /h            /? ;

:ABOUT          /? ;


: TFORTH      \ <> --- <>
        finduse
        doreset
        doboot
\        reset-options
;
                DEPRIVE

: doit
  S" /p 3 /s 2 /r /b /f twin.x tforth" EVALUATE
;
                                ( * End of File * )
