\ COPYRIGHT 1992@1997 DUTCH FORTH WORKSHOP - FILENAME : GETLINK.FRT #1


REVISION -getlink  " 1.0E --- getlink : INIT-CURSOR called      ---"

\ ******************************************************************************
\               Variables used also in display
\ ******************************************************************************

VARIABLE RECORD-ATTEMPT       \ We are trying to do a record
2VARIABLE CURRENT-RECORD
2VARIABLE OUR-RECORD
100.000.000.000 2CONSTANT CHUNKS        \ Chunks to save on disk

0 RECORD-ATTEMPT !
137.000.000.000.000 CURRENT-RECORD 2!
150.000.000.000.000 OUR-RECORD 2!

\ Limits for sieving in the current batch.
2VARIABLE N-START             \ inclusive
2VARIABLE CURRENT-START       \ is increased while we go
2VARIABLE N-END               \ Exclusive

\ Is there still work to be dealt for the transputers?
:  ?DONE         ( -- )
   N-END 2@ CURRENT-START 2@ D=
;

\ Number of a machine within a chain.
\ Applies to the latest debug or error messages.
VARIABLE MACHINE-ID
\ Current input channel, where the latest of the messages arrived.
0 VALUE N-CHANNEL

2VARIABLE TOTAL-TWINS           \ Total number of twins found, thus far
0 VALUE #PENDING                \ Number of chains of transputers still working

VARIABLE N-TWIN-OUT     \ Number of twin primes in latest batch
VARIABLE DEBUG-RESULT   \ The body of a debugging message

CHANNEL PTC     \ Our command channel

\ The length of the sieve in ENTRY 's
\ This is extremely short for testing
\ Lateron we will need three length's, for each length of transputer memory.
\ For convenience with replicate the PTC channel is put in the configuration
\ too, as entry zero
\ The link between is a link to a separate process that temporarily
\ takes over the output link, for sending long messages, in particular
\ the prime buffers. Normally there is a special process waiting on
\ LINK-BETWEEN. Only in that situation can the main process (getit)
\ use LINK-OUT
CREATE CONFIGURATION 5 4 * CELLS ALLOT
1 VALUE #LINKS     \ Includes command
CONFIGURATION VALUE CA  \ COnfiguration allocator
: CA, CA ! 1 CELLS +TO CA ;
: RESET-CONFIG
CONFIGURATION TO CA  \ COnfiguration allocator
\ LINK IN    LINK OUT    LINK-BETWEEN N-MACHINE   MEM (WORDS)
   PTC      CA,     0     CA,    MININT  CA,  0 CA,           0 CA,
   1 TO #LINKS
;
: ADD-CONFIG  ( link-nr, length, megs )
    ROT CELLS   DUP LINK0-IN + CA,  LINK0-OUT + CA,
    MININT CA,   SWAP CA, 1000000 4 / * CA,
    1 +TO #LINKS
;

\ Returns the addres of configuration #N ( int N -- address A)
: CONFIGURATION[] 5 * CELLS CONFIGURATION + ;
\ Content of a field of a configuration  ( address C -- int F )
: >LINK-IN   0 CELLS + @ ;
: >LINK-OUT  1 CELLS + @ ;
: >LINK-BETWEEN  2 CELLS + ;    \ No fetch! This IS the channel
: >N-MACHINE     3 CELLS + @ ;
: >MEM           4 CELLS + @ ;

\ ******************************************************************************
\               Here start the communication part
\ ******************************************************************************

NEEDS -twinmisc
NEEDS -logtwinc
NEEDS -display
NEEDS -twincon

: RESET-CHANNELS-IN-USE
    PTC RESET-CHANNEL DROP
    #LINKS 0 DO
        I CONFIGURATION[]
        DUP >LINK-IN RESET-CHANNEL DROP
        DUP >LINK-OUT RESET-CHANNEL DROP
            >LINK-BETWEEN RESET-CHANNEL DROP
   LOOP
;

\ Boot all channels with the default options that should have been set
: BOOT-ALL
    RESET-CHANNELS-IN-USE
    #LINKS 1 DO
        I CONFIGURATION[] >LINK-OUT MININT - 4 / TO port
        TFORTH
    LOOP
;
\ Gives the amount of numbers that can be done in one batch
\ on the transputers hanging off channel CHAN
: MAX-BATCH     ( int CHAN -- double LEN )
    CONFIGURATION[]
    DUP >MEM SWAP   >N-MACHINE NPE *   M*
;

?DEF DEBUG [IF]
    CR .( Expect the same values on two consequtive line )
    CR 1 CONFIGURATION[] >LINK-IN   H. SPACE
    1 CONFIGURATION[] >LINK-OUT H. SPACE
    1 CONFIGURATION[] >N-MACHINE .
    1 CONFIGURATION[] >MEM       .
    1 MAX-BATCH                  D.
    CR LINK3-IN H.
    SPACE LINK3-OUT H.
    SPACE 16 .
    250000 .
    SPACE 840000000 .
[THEN]

VARIABLE PrimesOffset1  \ offset1 of prime batch
LINK3-IN VALUE USE-IN   \ Currently selected channel (in)
LINK3-OUT VALUE USE-OUT \ Currently selected channel (out)
VARIABLE N-MACHINE-BACK \ Return message of startup, has gone through the
                        \ whole chain and back.
VARIABLE PRIMES-OFFSET  \ The offset of the primes buffer requested

\ Check whether the auxiliary process is not currently using
\ the output channel, i.e. it listens to us.
: CHECK-CHANNEL
    N-CHANNEL CONFIGURATION[] >LINK-BETWEEN @ MININT = IF
        ." Output channel is engaged: " N-CHANNEL .
        1000 MS \ Does it help?  This waiting.
        N-CHANNEL CONFIGURATION[] >LINK-BETWEEN @ MININT = IF
            STOP
        THEN
    THEN
;
\ All these communication use the currently selected channel
: MS->   CHECK-CHANNEL USE-OUT CHANNEL-! ;
: WS->   CHECK-CHANNEL USE-OUT CHANNEL-! ;
: DS->   CHECK-CHANNEL SWAP   USE-OUT CHANNEL-! USE-OUT CHANNEL-!  ;
: $S->   CHECK-CHANNEL USE-OUT CHANNEL-SEND ;
: $R<-   USE-IN CHANNEL-RECEIVE ;
: MR<-   USE-IN CHANNEL-@  DUP DEBUG-RESULT ! <> IF 0 1001 .LOG-ERROR ENDIF ;
: WR<-   USE-IN CHANNEL-@  SWAP ! ;
: DR<-   USE-IN CHANNEL-@  USE-IN CHANNEL-@  ROT 2! ;


\ ##################################################################
\ ####### Communication framework for a single processor ###########
\ ##################################################################

VARIABLE FORTHMessage
\ Start up the chains: send the first program its sieve length
\ and the chain length such that it can boot and send itself to
\ all other transputers in the chain
: STARTUP  ( -- )
    #LINKS 1 DO
        I TO N-CHANNEL \ To get error messages right
        I CONFIGURATION[]
        DUP >LINK-IN TO USE-IN
        DUP >LINK-OUT TO USE-OUT
        %>init   MS->
        DUP >MEM NPE * WS->
        DUP >N-MACHINE WS->
        %>init   MR<-
        N-MACHINE-BACK WR<-
        1 N-MACHINE-BACK +!
        DUP >N-MACHINE N-MACHINE-BACK @ = IF
              {{ CR N-MACHINE-BACK ? ." transputers successfully booted" }}
        ELSE
            {{ CR N-MACHINE-BACK ? ." N-MACHINE check went wrong: " }}
        THEN
        DROP
    LOOP
;


\ Asks the transputers on channel CHAN of the configuration to
\ do the maximum amount of work from CURRENT-START.
: SIEVE-COMMAND   ( int CHAN -- )
    LOCAL CHAN

    CHAN TO N-CHANNEL
    CHAN CONFIGURATION[] >LINK-IN   TO USE-IN
    CHAN CONFIGURATION[] >LINK-OUT  TO USE-OUT
    %>start          MS->
    CURRENT-START 2@ DS->
    CHAN MAX-BATCH   N-END 2@ CURRENT-START 2@ D- DMIN    ( -- work )
    2DUP              DS->
    CURRENT-START D+!
    MASK-IN-MAIN NPE CELLS $S->
    1 +TO #PENDING
;


: OFFICIAL-RUN

    TIMER-RESET
    N-START 2@ D0= IF
        \ SKip over the first part with too small primes.
        N-SKIP S>D CURRENT-START 2!
        TWINS-SKIPPED S>D TOTAL-TWINS 2!
    ELSE N-START 2@ N-SKIP S>D D< IF
         N-START D>S 1003 .LOG-ERROR
    ELSE
        0.0 TOTAL-TWINS 2!
        N-START 2@ CURRENT-START 2!
    THEN THEN
    CURRENT-START 2@ .STATUS
    #LINKS 1 DO
        ?DONE 0= IF
            I SIEVE-COMMAND
        THEN
    LOOP
;

: WRITE-RESULT?
        \ Claim and or wait for connextion
        N-START 2@
        N-END 2@
        TOTAL-TWINS 2@
        WRITE-RESULT
        N-END 2@ N-START 2!
        N-START 2@ CHUNKS D+ N-END 2!
        N-START 2@ OUR-RECORD 2@ D< IF
            OFFICIAL-RUN
            START-MESSAGE
        THEN
;

\ A new result has been obtained. Accumulate it into the total and
\ restart new processes if necessary.
: ACCUMULATE-TWIN-COUNT
    N-CHANNEL ERASE-STARTS
    -1 +TO #PENDING
    N-TWIN-OUT @ S>D TOTAL-TWINS D+!
    .LAST-ACCUMULATED
    CURRENT-START 2@    \ Remember before we increment it.
    ?DONE 0= IF
        N-CHANNEL SIEVE-COMMAND
    THEN
    .STATUS
    #PENDING 0= RECORD-ATTEMPT @ AND IF
        WRITE-RESULT?
    THEN
;

FORTH-PROCESS SEND-PRIMES       \ Three parallel processes
: SEND-PRIMES-BODY
    PAR
    #LINKS 1- 0 TIMES
    I 1 :I
    STARTPROCESS
        1+ >R   \ My channel number remains on the return stack forever
        BEGIN
            R@ CONFIGURATION[] >LINK-BETWEEN CHANNEL-@  ( -- offset )
            DUP %>stop = IF STOP THEN
            \ Send message idea
            %>primebatch  R@ CONFIGURATION[] >LINK-OUT CHANNEL-!
            \ Send offset, keep it
            DUP R@ CONFIGURATION[] >LINK-OUT CHANNEL-!
            \ Get bufferstart, length
            PPBY / SIEVE + BYTES-PER-PRIME-BATCH
            R@ CONFIGURATION[] >LINK-OUT CHANNEL-SEND   \ Send buffer
        AGAIN
    ENDPROCESS
    ENDPAR
;


\ Control centre: receive all messages, react to them, accumulate
\ results and maintain the information in behalf of display.
\
FORTH-PROCESS TWIN-CONTROL
: TWIN-CONTROL-BODY
    1000 MS STARTUP

    BEGIN
        \ Wait till somebody sends us something
        SELECT
           #LINKS 0 REPLICATE I CONFIGURATION[] >LINK-IN GUARD ENDGUARD
        ENDSELECT

        TO N-CHANNEL
        N-CHANNEL CONFIGURATION[] >LINK-IN    TO USE-IN
        N-CHANNEL CONFIGURATION[] >LINK-OUT   TO USE-OUT

        \ Decide based on the message identification
        FORTHMessage WR<-
        FORTHMessage @ CASE

            %>stop OF
                #LINKS 1 DO
                    %>stop I CONFIGURATION[] >LINK-BETWEEN CHANNEL-!
                LOOP
                STOP
            ENDOF

            %>start OF
                OFFICIAL-RUN
                START-MESSAGE
            ENDOF

            %<requestbatch    OF
                PRIMES-OFFSET WR<- \ offset1 of prime batch
                PRIMES-OFFSET @
                DUP N-CHANNEL   REMEMBER-PRIMES
                N-CHANNEL CONFIGURATION[] >LINK-BETWEEN CHANNEL-!
            ENDOF

            %<releasebatch    OF
                PRIMES-OFFSET WR<- \ offset1 of prime batch
                \ Ignore for the moment.
            ENDOF

            %<twin OF
                N-TWIN-OUT WR<-
                ACCUMULATE-TWIN-COUNT
            ENDOF

            %>ping OF
                #LINKS 1 DO
                    I TO N-CHANNEL \ To get a correct check
                    I CONFIGURATION[] >LINK-OUT TO USE-OUT
                    %>ping MS->
                LOOP
            ENDOF

            %<debug OF
                MACHINE-ID WR<-   DEBUG-RESULT WR<-
                {{ 0 #LL AT-XY ." DEBUG " .MACHINE-ID DEBUG-RESULT @ DUP . H. SPACE }}
            ENDOF

            %<showstart OF
                MACHINE-ID WR<-   DEBUG-RESULT DR<-
                DEBUG-RESULT 2@   MACHINE-ID @   N-CHANNEL   REMEMBER-START
            ENDOF

            %<error OF
                MACHINE-ID WR<-   DEBUG-RESULT WR<-
                DEBUG-RESULT @ MACHINE-ID @ .LOG-ERROR
            ENDOF

            {{ 0 #LL AT-XY ." unofficial DEBUG/spurious message " DUP . H. SPACE }}
        ENDCASE
    AGAIN
;


: STARTIT %>start PTC CHANNEL-! ;
: STOPIT %>stop PTC CHANNEL-! ;
\ Calculate the primes up till UPLIM
: CALC  ( double UPLIM -- )
    N-END 2!
    0.0 N-START 2!
    INIT-CURSOR
    STARTIT
;

: CALC-RANGE  ( double DOWNLIM double UPLIM -- )
    N-END 2!
    N-START 2!
    INIT-CURSOR
    STARTIT
;


\ Try to read last line from result.txt into LINE.

CREATE LINE 129 ALLOT
LINE 1+ CONSTANT LINE-BUFFER
: GET-LAST-LINE
    0 LOCAL COUNT         \ Count of LINE-BUFFER characters
      S" RESULT.TXT" R/O OPEN-FILE THROW
    LOCAL FILE
    0.0 0.0     \ Remember last two file-positions
    BEGIN
        FILE FILE-POSITION THROW
        2ROT 2DROP      \ Get rid of ante-paenultima file-position
        LINE-BUFFER 126 FILE READ-LINE THROW DROP
    ( eof) 0= UNTIL
    2DROP         \ Drop unsuccessful position
    ( last sucessful position) FILE REPOSITION-FILE THROW
     LINE-BUFFER 126 FILE READ-LINE THROW 0= THROW
    ( count) LINE C!
    FILE CLOSE-FILE THROW
;

\ Finds the point where to continue for the record, based on the result
\ file.
: FIND-LAST     ( -- double D )
    ['] GET-LAST-LINE CATCH ABORT" Could get last line of results file"
    [CHAR] . LINE $+C!
    LINE $@ EVALUATE
    DLOCAL CHECKSUM
    DLOCAL RESULT
    DLOCAL HIGH
    DLOCAL LOW
    LOW HIGH RESULT    D+ D+    CHECKSUM D=  0= ABORT" Checksum failure"
    HIGH
;
: DO-RECORD  ( double DOWNLIM -- )
    RECORD-ATTEMPT ON
    FIND-LAST N-START 2!
    N-START 2@ CHUNKS D+ N-END 2!
    STARTIT
;
