
\ This module hides the presentation of the primes in
\ `CalcPrimes ->Buffer'
\ It needs the externals `CalcPrimes ->Buffer' `CalcPrimes ->Offset' `p' .
\ It exposes only `SieveOneBatch' and `MASK'

     REVISION -mask " MASK 's 0.3V Finding of primes itself here "

PRIVATES

DOC
Upper limit 1000
7 , \ 0
9 , \ 1
19 , \ 2
21 , \ 3
31 , \ 4
33 , \ 5
37 , \ 6
39 , \ 7
49 , \ 8
51 , \ 9
61 , \ 10
63 , \ 11
67 , \ 12
69 , \ 13
79 , \ 14
81 , \ 15
91 , \ 16
93 , \ 17
109 , \ 18
111 , \ 19
121 , \ 20
123 , \ 21
151 , \ 22
153 , \ 23
157 , \ 24
159 , \ 25
187 , \ 26
189 , \ 27
199 , \ 28
201 , \ 29
ENDDOC

WP @ CONSTANT SMASK PRIVATE     \ Fake an address in the initialised workspace
HEX
0 W,  01 W,  0 W, 0 W, 0 W, 0 W, 0 W,  02 W,
0 W, 0 W, 0 W,  04 W,  0 W,  08 W,  0 W, 0 W,
0 W,  10 W,  0 W,  20 W,  0 W, 0 W, 0 W,  40 W,
0 W, 0 W, 0 W, 0 W, 0 W,  80 W,
DECIMAL

SMASK 30 + WP @ <> [IF] .( Incorrect #Entries in SMASK) ABORT [THEN]

FWORD PointerWalk  PRIVATE      \ Into the CalcPrimes ->Buffer
FWORD PrimesMask   PRIVATE      \ Mask currently in use over CalcPrimes ->Buffer
FWORD CounterM     PRIVATE      \ Counts the masks beings tried
\ These pointers are in behalf of clients that need to walk a part of
\ `CalcPrimes ->Buffer'
FWORD PrimesPointerLow  \ Address of lowest byte
FWORD PrimesPointerHigh \ Address of lowest byte

\ THe following being procs favors them being placed in low memory
\ Execute the sieving action for large primes for the primes
\ between PointerWalk en PrimesPointerHigh found by PrimesMask.
\ N.B. PointerWalk is destroyed.
PROC DoLargePrimes
    BEGIN,
        PointerWalk ldl, PrimesPointerHigh ldl, <, WHILE,
        PointerWalk ldl,  lb,
        PrimesMask ldl, and,   IF,
            remove-large-prime
        THEN,
        p ldl, NPE adc, p stl,
        PointerWalk ldl, NPE PPBY / adc, PointerWalk stl,
    REPEAT,
END-PROC

\ Execute the sieving action for small primes for the primes
\ between PointerWalk en PrimesPointerHigh found by PrimesMask.
\ N.B. PointerWalk is destroyed.
PROC DoSmallPrimes
    BEGIN,
        PointerWalk ldl, PrimesPointerHigh ldl, <, WHILE,
        PointerWalk ldl,  lb,
        PrimesMask ldl, and,   IF,
            remove-small-prime
        THEN,
        p ldl, NPE adc, p stl,
        PointerWalk ldl, NPE PPBY / adc, PointerWalk stl,
    REPEAT,
END-PROC

\ Pass through the batch of primes from register B to register A
\ evaluates string S for all primes found. Presumably this amounts
\ to a sieving action.
\ The auxiliary table of masks (SMASK) contains byte, like the
\ CalcPrimes ->Buffer, but there we need not use the more expensive byte
\ addressing.
MACRO SieveOneBatch  ( int I -- )
    LOCAL Command      \ Prevent interference with run time stack
\         CalcPrimes ->Offset ldl, DEBUG,
\         i-p-s ldl, DEBUG,

    ( reg A) CalcPrimes ->Buffer ldl, bsub,   PrimesPointerHigh stl,
    ( reg B) CalcPrimes ->Buffer ldl, bsub,   PrimesPointerLow  stl,
    \ We take together all primes with the same remainder (mod NPE)
    \ so we go NPE times through CalcPrimes ->Buffer.
    \ Fortunately a lot of remainders do not lead to prime numbers.
    0 ldc, CounterM stl,
    the-sieve ldl, UpLim stl,     \ For the inner loops to profit
                                    \ from fast variables.
    BEGIN,
        \ The prime currently considered is the prime belonging to the
        \ start of the table, plus PPBY for each byte from that start
        \ plus the remainder we work with currently
        CalcPrimes ->Offset ldl,
        PrimesPointerLow  ldl, CalcPrimes ->Buffer ldl, sub, 
        PPBY ldc, prod, add,
        CounterM ldl, add,
        p stl,

        \ The mask belonging to this remainder
        CounterM ldl,   PPBY ldc, rem, SMASK ldlp, wsub, 0 ldnl,
        PrimesMask stl,

        PrimesMask ldl, IF,    \ Bit set in MaskBuffer?
        p ldl, 7 ldc, rem, IF,  \ "Primes" divisible by 7 play havoc in `Learn'
            \ At this point `p' may not be a prime, but it is relative prime
            \ to 210, so it is usable in Learn.
            Learn   \ How to handle primes with this remainder (mod NPE).
            PrimesPointerLow ldl,   CounterM ldl, PPBY ldc, div,   add,
            PointerWalk stl,
            \ remove prime found using PrimesMask
            Command 1 = IF DoSmallPrimes THEN
            Command 2 = IF DoLargePrimes THEN
        THEN,
        THEN,

        CounterM ldl, 1 adc, CounterM stl,
    CounterM ldl, NPE 1- ldc, >, UNTIL,

END-MACRO
DEPRIVE
