\ COPYRIGHT 1992@2000 DUTCH FORTH WORKSHOP - FILENAME : TWIN.FRT #0

             NEEDS -asswrite

REVISION  -twin "--- 2.0B Twin Main : PrimeLen used ---"

\ LAST CHANGE: at the end of the range one entry too many is Erase

\ ******************************************************************
\ This program is intended to run on a chain of transputers with
\ the same memory, all booting from link0 and connected via
\ link2 to the next.
\ It contains of a communication of a calculation and a communication
\ process. This source contains the calculation. The calculation recieves
\ a command to sieve a range of numbers and counts the primes in it.
\ It issues requests for batches of primes, that are used to eliminate
\ primes. The sieve has a data structure specific to twin prime counting.
\ This data structure imposes a peculiar way of sieving.

\ NOTIONS:
\ A "calculation" is the sieving and reporting of the twin in a range
\ of numbers.
\ A "batch" is the sieving of such range using a batch of primes.
\ An "entry" is a 32-bits words representing 210 numbers, 30 of them
\ actually are represented by a bit.
\ A "number index" is an index counting numbers from the start of the
\ sieve. It may not have a bit in the table.
\ An "entry index" is an index counting entry's from the start of the
\ table.

\ ************** constants for asswrite ****************************
\ Start of the area where the code is stored as a result of loading
\ the program. This need not be the execution start address
$80000A00 TO CodeStart

\ NOTE: for comparisons it is important that there is an
\ enormous gap between the end of the sieve and boundary of the memory
\ This can be accomplished by reversing the sieve and starting from the
\ highest multiples. If the lowest part of the sieve is passed, the
\ sieving can finish.
\ (It may be that sieving up works equally well with an unreversed sieve
\ but I do not seem to get into the right mindset for that.).
\ Worst case a next lower addres is 28*(p/NPE) words lower than the
\ previous one. This must not skip over the whole (HeapStart, maxint) range.
\ (Lest it becomes lower that Heapstart and the sieving will result in setting
\ random bits in the memory allocated to the program.)
\ As NPE is at least 210 this doesnot restrict this program because it
\ allows primes greater than 32 bits. But these would require rewriting the
\ program anyway.
\ Fall 1999 the world record is between 10^15 and 10^16.
\ This probably is out of reach, not for the program but for the computing
\ power needed.
$80002400 TO HeapStart

INIT-ASSEMBLER                  \ Get changes accross to the assembler

     NEEDS -twincon
     NEEDS -genmac
     NEEDS -primtab
     NEEDS -commun
NUMBERS-PER-ENTRY HARRAY MASK-COPY

PRIME-BUFFER CalcPrimes

\ #################### Global data ###########################################


\ ******************** arrays ************************************************
20 HARRAY COMM-WS-UP
20 HARRAY COMM-WS-DOWN

\ NOTE: this MUST be the last HARRAY, if the program must work
\ for different values of memory.
\ So this is a forward reference, hence a zero length.
\ The actual length is x + 2 words , where x is 10^6 time the
\ number of megawords in the machine. (a megaword is 2^12 bytes )
\ LEN-SIEVE is an attempt to define x for a 4 Mbyte machine.
\ The two extra words are needed because we shift the buffer slightly
\ such that MAGIC-OFFSET condition is met, see below.
\ The unused numbers represented in the buffer are screened of by marking
\ them as non-prime, see below.
\ Forward reference to ACTUAL-SIEVE, use as if it was a HARRAY
0 VALUE HighPointOfSieve
\
\ The following numbers define the internal structure of the-sieve for
\ one calculation.
\
\ The number such that the first ENTRY from ``the-sieve'' represents the
\ numbers NumberStart+linverse[0], NumberStart+linverse[1] etc.
\ The next entry is lower in memory and represents NPE+NumberStart+linverse[0] etc.
\ Only one seventh of the numbers are represented. This is crucial
\ for the compactness and speed of the program.
\ For this to work NumberStart modulo NPE should be MAGIC-OFFSET.
\ WARNING: while starting these variables temporarily contain
\ data different from the above description.
\ The interval [ NumberStart, NumberStart+NumberLen) is to be sieved
\ WARNING: these numbers must be even, but this is not asserted.
0. DWORD NumberStart
0. DWORD NumberLen              \ Counts numbers.

\ ********************** important variables *******************************

\ The following variables are important because:
\    you must understand them to understand the program
\    they are accessed very often, so they need in fast ram.
\ By defining them here the only faster variables are the scratch
\ variables in GENMAC.FRT that are accessed very often too.

FWORD p         \ The current prime being used for elimination
FWORD q         \ The current prime divided by NPE, this is approximately
                \ the number of entry's skipped to arrive at the next
                \ multiple of p.
FWORD m         \ The current mask used to eliminate primes by anding.

\ Last number offset in SIEVE[0..NumberLen] that represents a multiple of p.
\ There may be no actual bit available for this number in the table !
\ This can point 1 bit outside of the table.
\ If Numberlen where odd, this could actually represent a number.
FWORD Start

\ Points to the entry in the sieve containing the lowest numbers.
\ For practical purposes it is like a HARRAY but indexes down.
FWORD the-sieve


\ ******************************************************************************
\ The following numbers are constant for a calculation
\ ******************************************************************************

0. DWORD NumberEnd       \ NumberStart+NumberLen
\ Size of the sieve in numbers sieved,
\ this size is forced upon us by the host program.
\ It must be smaller than NPE times actual memory size in words.
\ Because of big endianness we can use it as a FWORD, whenever we want to.
0. DWORD SieveSize

\ Size of the sieve in entries, i.e. the-sieve[EntryLen]
\ is one pas the last bit representing a number to be sieved.
\ That bit belongs to the number NumberStart+NumberLen-1
\ The sieve itself has the lowest indices at the higher addresses
FWORD EntryLen

MACRO TDEBUG,
      EntryLen ldl, DEBUG,
      0 ldc,  the-sieve ldl, wsub, 0 ldnl, DEBUG,
      -1 ldc, the-sieve ldl, wsub, 0 ldnl, DEBUG,
      -2 ldc, the-sieve ldl, wsub, 0 ldnl, DEBUG,
END-MACRO

0 WWORD #EmptyEntries   \ Padding up front the table.

\ ******************************************************************************
\ The following numbers are constant for a batch
\ ******************************************************************************

\ Byte indices pointing into CalcPrimes ->Buffer:
FWORD i-p-small    \ Start of sieving by small prime method
FWORD i-p-large    \ Start of sieving by large prime method
FWORD i-p-sqrt     \ Start of no-sieving (no need to go beyond p-sqrt ).

\ ******************************************************************************
\ Macros to setup numbers for a calculation
\ ******************************************************************************

PRIVATES

\ p-sqrt is the largest prime that has to be sieved in the current calculation
\ Initialise p-sqrt to a reasonable starting value
\ It is left at its current value for the next calculation
\ In the long run that is very fast.
\ tested
1000 WWORD p-sqrt PRIVATE

\ NumberStart[0, NumberLen-1] must be sieved.
\ A fine point is handled in this MACRO.
\ We need to sieve two more entries to catch a twin prime that ends
\ with one of the first two numbers.
\ Such a primetwin was not counted with the preceeding batch (!).
\ tested
MACRO AdjustCommandInput
      2 ldc, Temp stl,
      NumberStart Temp DS-, NumberStart dstl,
      NumberLen ldl,   2 ldc, add,   NumberLen stl,
END-MACRO


\ NumberStart must have the same remainder modulo NPE as MAGIC-OFFSET
\ We accomplish this by adding an offset to NumberStart and NumberLen
\ called #EmptyEntries . This number of entries has to be zeroed in
\ the table, such that they are not counted as primes.
\ tested
MACRO FitToTable
    NPE ldc, Temp stl,
    NumberStart dldl, Temp lrem, NPE adc,
    MAGIC-OFFSET ldc, sub,   \ Compare with the base for the tables
    NPE ldc, rem, #EmptyEntries stl,
    NumberStart #EmptyEntries DS-, NumberStart dstl,
    NumberLen ldl,   #EmptyEntries ldl,   add,   NumberLen stl,
END-MACRO PRIVATE


\ Fill p-sqrt with the square root of the largest number in the table.
\ It is well known that we need not sieve beyond that.
\ Use Newton. Keep a safety margin of 2, because it is approximate.
0 WWORD SqrtCompare PRIVATE
MACRO fill-p-sqrt
    $70000000 ldc, p-sqrt stl,
    BEGIN,
        NumberEnd dldl, rev,   p-sqrt ldl,   ldiv, \ Quotient
        p-sqrt ldl,
        dup, SqrtCompare stl,    \ Keep copy
        add, 1 ldc, shr,
        dup, p-sqrt stl, \ New value
        SqrtCompare ldl, sub, abs, 2 ldc, >, WHILE,
    REPEAT,
    p-sqrt ldl,   2 adc, p-sqrt stl,
END-MACRO PRIVATE

\ Prepare all the numbers that have a constant meaning through one
\ sieve action (value pair NumberStart/NumberLen):
\ the pair itself, NumberEnd EntryLen p-sqrt.
\ Initialise CalcPrimes.
MACRO CalculationSetup
    AdjustCommandInput
    FitToTable
    NumberStart NumberLen DS+,    NumberEnd dstl,
    NumberLen ldl,    NPE ldc,    div,   1 adc, EntryLen stl,
    EntryLen ldl, -1 adc, HighPointOfSieve ldl, wsub, the-sieve stl,
    fill-p-sqrt
    0 ldc, CalcPrimes ->Offset stl,
\     0 ldc, CalcPrimes ->Buffer stl,                     \ BUG??????
END-MACRO

\ ************************************************************************
\ Macros to setup numbers for a batch
\ ************************************************************************

\ Find the indices for separation in small and large method.
\ Only in the first batch of primes makes it sense to use the small method.
\ Also in the first batch skip byte zero with the primes <30 that are done
\ by wheels.
MACRO fill-i-p-limits

    CalcPrimes ->Offset ldl, 0 ldc, =, IF,
        1 ldc, i-p-small stl,
        \ Something clever related to the sieve length
        1000 ldc, i-p-large stl,
    ELSE,
        0 ldc, i-p-small stl,
        0 ldc, i-p-large stl,
    THEN,

END-MACRO PRIVATE

\ Find the index beyond which sieving is no longer needed.
MACRO fill-i-p-sqrt
    p-sqrt ldl,   CalcPrimes ->Offset ldl,   sub,
    PPBY ldc,   div,   1 adc, i-p-sqrt stl,
END-MACRO PRIVATE

\ Limit both above indices to pointing in the table
\ or -1 if the table is beyond the index.
FWORD Limit PRIVATE     \ Auxiliary, in order to use max, / min, macro's
MACRO LimitIndices

    0 ldc,   Limit stl,
    Limit i-p-sqrt max, i-p-sqrt stl,
    Limit i-p-large max, i-p-large stl,

    BPPB ldc,   Limit stl,  \ These limits are exclusive!
    Limit i-p-sqrt min, i-p-sqrt stl,
    Limit i-p-large min, i-p-large stl,

    i-p-sqrt i-p-large min, i-p-large stl,

END-MACRO PRIVATE

\ Set up everything for sieving a batch of primes
MACRO BatchSetup
    fill-i-p-limits
    fill-i-p-sqrt
    LimitIndices
END-MACRO

\ Make sure the data of the batch is present.
\ Wait for the communication to finish if need be.
MACRO WaitForBatchPresence

    %<requestbatch MSM<-P,   CalcPrimes ->Offset WSM<-P,
    %>primebatch MRM->P, CalcPrimes ->Offset WRM->P,
    CalcPrimes ->Buffer PrimeLen $RM->P,

END-MACRO
DEPRIVE

\ #################### sieve and removing ###########################################

\ CRITICAL ROUTINE.
\ May result in an index outside the table if NumberEnd is odd.
\ We cannot afford to assert such here for speed reasons.
\ Leaves into A the number-index into the-sieve of the last multiple
\ of the prime
\ tested & confirmed
MACRO last-p-mult       ( -- reg A)
    NumberEnd dldl,  p lrem,  NumberLen ldl, rev, sub, Start stl,
END-MACRO

\ #################### Needs global data ###################################

        NEEDS -learn
        NEEDS -removep
        NEEDS -mask
        NEEDS -commun1

\ All other space being allocated we can now allocate the huge sieve.
LEN-SIEVE HARRAY actual-sieve
\ We are in time to patch up HighPointOfSieve , because it has not been used
\ to generated code, it is only refered to in macro's
actual-sieve TO HighPointOfSieve

\ *************** finishing off the sieve **********************
\ Must be done after the wheels.

PRIVATES
\ reset the bit as with number index ni
\ tested
MACRO reset-bit  ( FWORD ni -- )
      DUP ldl,   NPE ldc, div, Temp stl,        \ Entry index
      ( index) ldl, NPE ldc, rem,               \ Mask index
      MASK-COPY ldl, wsub, 0 ldnl, \ Get mask
      Temp ldl, neg, the-sieve ldl, wsub, 0 ldnl,
      and,
      Temp ldl, neg, the-sieve ldl, wsub, 0 stnl,
END-MACRO PRIVATE


\ Erase the entries that must not be counted at the beginning
\ Erase up to but not including #EmptyEntries
\ tested
MACRO ErasePaddingAtBeginning
    #EmptyEntries ldl, -1 adc,  0 ldc,   UPDO,
        II reset-bit
    1 FIXED+LOOP,
END-MACRO PRIVATE

\ Erase the entries that must not be counted at the end
\ Start with NumberLen+1 and end at first possible boundary.
\ tested
MACRO ErasePaddingAtEnd
    \ Next higher multiple of NPE pas NumberLen.
    \ Has to be calculated first, because it needs all 3 registers.
    NumberLen ldl, NPE adc, dup, NPE ldc, rem, sub, -1 adc,
    NumberLen ldl, 1 adc,
    UPDO,
        II reset-bit
    1 FIXED+LOOP,
END-MACRO PRIVATE

MACRO ErasePadding
        ErasePaddingAtBeginning
        ErasePaddingAtEnd
END-MACRO

DEPRIVE

\ *************** sieving by wheels ************************************

PRIVATES

0 WWORD OurEntryLen  PRIVATE   \ The EntryLen while with ``remove-with-wheel''
0 WWORD LenWheel     PRIVATE   \ Current length of the wheel
0 WWORD Pointer1     PRIVATE   \ Source pointer
0 WWORD Pointer2     PRIVATE   \ Upper Limit

\ Make the sieve larger by a factor P (but not larger than ``EntryLen''
\ entries) replicating the information of divisibility by smaller primes.
\ Remove the multiples of P from the resulting sieve.
\ This a PROC because it is called many times.
\ WARNING: in this macro NumberStart and NumberLen are temporarily changed.
PROC remove-with-wheel   ( int P -- )
    p stl,
    LenWheel ldl,   p ldl,   prod,  Temp stl,
    EntryLen Temp min,   OurEntryLen stl,

    \ This condition prevents overwriting of the-sieve[EntryLen]
    \ (We are going to use a UNTIL, loop!).
    LenWheel ldl, OurEntryLen ldl, <>, IF,
        \ Now replicate
        the-sieve ldl,   Pointer1 stl,
        LenWheel ldl, OurEntryLen ldl, sub,   ( -[p-1]*LenWheel )
        1 adc,         \ Change from "one past" to "at" in order to be
                        \ able to use <, instead of <=,
        the-sieve ldl, wsub,   Pointer2 stl,     \ Last entry created.

        \ This replication is including the upper limit, because in the last time
        \ we must set the-sieve[EntryLen] that contains some bits.
        BEGIN,
            Pointer1 ldl, 0 ldnl,    \ Fetch one entry
            LenWheel ldl, neg, Pointer1 ldl, wsub,   0 stnl, \ Next wheels impression
            Pointer1 ldl, -4 adc, Pointer1 stl,
        Pointer1 ldl, Pointer2 ldl, <, UNTIL,
    ENDIF,

    \ Now make the outside world do their job of removing
    OurEntryLen ldl, NPE ldc, mul,   NumberLen stl,
    NumberStart NumberLen DS+,    NumberEnd dstl,
    remove-wheel-prime

    OurEntryLen ldl,   LenWheel stl,

END-PROC PRIVATE

0 WWORD NumberLenCopy PRIVATE
0 WWORD NumberEndCopy
\ Fill the sieve such that multiple of primes < 30 are crossed or not
\ represented.
\ Temporarily fake the rest of the system (in fact ``last-p-mult'' and
\ ``remove-wheel-prime'' that ``NumberLen and NumberEnd'' are different
\ than what they actually are.
MACRO FillSieve ( -- )
    NumberLen ldl,   NumberLenCopy stl, \ Save old information
    NumberEnd ldl,   NumberEndCopy stl,

    \ Initialise wheel,
    TWIN-ALL-SET ldc,   the-sieve ldl, 0 stnl,
    1 ldc,   LenWheel stl,

    \ Use all the remaining primes <PPBY
    \ i.e. all those in the first byte of the first buffer
    11 ldc, remove-with-wheel
    13 ldc, remove-with-wheel
    17 ldc, remove-with-wheel
    19 ldc, remove-with-wheel
    23 ldc, remove-with-wheel
    29 ldc, remove-with-wheel

    OurEntryLen ldl, EntryLen ldl, <>, IF, 3 ERROR, ENDIF,

    NumberLenCopy ldl, NumberLen stl, \ Restore old information
    NumberEndCopy ldl, NumberEnd stl,
END-MACRO

DEPRIVE

\ #################### counting ###########################################

FWORD n-twin

\   Count the number of twinprimes in all numbers represented
\   in the table ``sieve[]'' with the usual coding.
\ tested
FWORD Pointer1    FWORD Pointer2
MACRO CountTwins ( -- )

    \ It is correct to handle whole ENTRY's here, see erase-extra-entry
    \ Also we can use the underlying HARRAY itself
    \ First entry
    0 ldc, HighPointOfSieve ldl, wsub,        Pointer1 stl,
    \ Last entry
    EntryLen ldl, -1 adc, HighPointOfSieve ldl, wsub,   Pointer2 stl,

    0 ldc,   n-twin stl,             \ Initialise twin count to zero
    Pointer1 ldl,
    0 ldc,      \ Trick explained in removep.frt
    BEGIN,
        diff,           \ Trick explained in removep.frt
        dup, 0 ldnl,           \ Fetch
        n-twin ldl,
        rev, bitcnt,            \ Accumulate count into n-twin
        n-twin stl,
        4 adc,
   dup, Pointer2 ldl, >, UNTIL,

END-MACRO


\ Remove all multiples of primes from SIEVE such that the primes remain
\ tested
MACRO DoSieve
    BEGIN,
        BatchSetup
    i-p-sqrt ldl, 0 ldc, >, WHILE,
        NumberStart 1+ ldl, 0=, IF,     \ No m.s. part
        NumberStart ldl, 0 ldc, >, IF,  \ L.s. part is unsigned < 2G
            CalcPrimes ->Offset ldl,   i-p-sqrt ldl, PPBY ldc, mul, sum,
            NumberStart ldl, >, IF,
                 \ Prime itself is present in the table, so it could sieve
                 \ itself away.
                 #32 ERROR,
            CalcPrimes ->Offset ldl,   i-p-sqrt ldl, PPBY ldc, mul, sum,
            DEBUG,
            THEN,
        THEN, THEN,

        WaitForBatchPresence

        i-p-small ldl,   i-p-large ldl,  1 SieveOneBatch
        i-p-large ldl,   i-p-sqrt ldl,   2 SieveOneBatch

        \ prepare for next batch
        CalcPrimes ->Offset ldl,   PrimeLenAdd,   CalcPrimes ->Offset stl,
    REPEAT,
END-MACRO


{
ENTER-HERE
    \ Initialise workspace to the 28 needed for the system and 5 for us.
    sethalterr,
    WorkSpaceStart ldc,  gajw,

    \ Boot through, initialise all systems down (after message to that effect)
    BootChain
    COMM-UP          \ Start the main communication process that sends up
    COMM-DOWN        \ Start the main communication process that sends down

    BEGIN,
        CheckCrc
        \ Synchronise with COMM
        %>start MRM->P, NumberStart DRM->P,   NumberLen DRM->P,
        %>notify MSM<-P,
        \ Send the actual start of this processor for display
        %<showstart MSM<-P,   ProcessorId WSM<-P, NumberStart DSM<-P,
        \ COMM has now filled `ProcessorId' and `NumberStart'
        CalculationSetup
        FillSieve
        ErasePadding
        DoSieve
        CountTwins
        CheckCrc

        %<twin MSM<-P,   n-twin WSM<-P,
    AGAIN,
}

AddBoot myboot.x
WriteOut twin.x

\ -twin
